from importlib.util import find_spec
import logging
from functools import wraps
from typing import Callable

log = logging.getLogger(__name__)


if find_spec("memory_profiler"):
    from memory_profiler import memory_usage
else:
    raise ImportError("Install memory_profiler.")


def _func_full_name(func: Callable):
    return getattr(func, "__qualname__", repr(func))


def mem_profile(func: Callable) -> Callable:
    """A function decorator which profiles the memory used when executing the
    function. The logged memory is collected by using the memory_profiler
    python module and includes memory used by children processes. The usage
    is collected by taking memory snapshots every 100ms. This decorator will
    only work with functions taking at least 0.5s to execute due to a bug in
    the memory_profiler python module. For more information about the bug,
    please see https://github.com/pythonprofilers/memory_profiler/issues/216

    Args:
        func: The function to be profiled.

    Returns:
        A wrapped function, which will execute the provided function and log
        its max memory usage upon completion.

    """

    @wraps(func)
    def with_memory(*args, **kwargs):
        log = logging.getLogger(__name__)
        mem_usage, result = memory_usage(
            (func, args, kwargs),
            interval=0.1,
            timeout=1,
            max_usage=True,
            retval=True,
            include_children=True,
        )
        # memory_profiler < 0.56.0 returns list instead of float
        mem_usage = mem_usage[0] if isinstance(mem_usage, (list, tuple)) else mem_usage
        log.info(
            "Running %r consumed %2.2fMiB memory at peak time",
            _func_full_name(func),
            mem_usage,
        )
        return result

    return with_memory



from importlib.util import find_spec

from .mlflow_utils import hook_impl, mlflow_log_metrics, mlflow_log_params


class MLflowDataSetsLoggerHook:
    """Logs datasets of (list of) float/int and str classes to MLflow"""

    _logged_set = set()

    def __init__(self, enable_mlflow: bool = True):
        """
        Args:
            enable_mlflow: Enable logging to MLflow.
        """
        self.enable_mlflow = find_spec("mlflow") and enable_mlflow

    @hook_impl
    def after_node_run(self, node, catalog, inputs, outputs):
        for name, value in inputs.items():
            if name not in self._logged_set:
                self._logged_set.add(name)
                self._log_dataset(name, value)

        for name, value in outputs.items():
            if name not in self._logged_set:
                self._logged_set.add(name)
                self._log_dataset(name, value)

    def _log_dataset(self, name, value):
        if isinstance(value, str):
            mlflow_log_params({name: value}, enable_mlflow=self.enable_mlflow)
        elif isinstance(value, (float, int)):
            mlflow_log_metrics({name: value}, enable_mlflow=self.enable_mlflow)
        elif isinstance(value, (list, tuple, set, dict)):
            mlflow_log_params(
                {name: "{}".format(value)[:100]}, enable_mlflow=self.enable_mlflow
            )


class MLflowOutputsLoggerHook(MLflowDataSetsLoggerHook):
    """Deprecated alias for `MLflowOutputsLoggerHook`"""



from importlib.util import find_spec
from logging import getLogger
import tempfile
from typing import Any, Dict, Union  # NOQA

from kedro.io import AbstractDataSet, DataCatalog
from kedro.pipeline.node import Node
from kedro.pipeline import Pipeline

from .mlflow_utils import (
    hook_impl,
    mlflow_log_artifacts,
    mlflow_log_metrics,
    mlflow_log_params,
    mlflow_log_values,
)


log = getLogger(__name__)


def get_kedro_runner():
    import inspect
    from kedro.runner import AbstractRunner

    return next(
        caller[0].f_locals.get("runner")
        for caller in inspect.stack()
        if isinstance(caller[0].f_locals.get("runner"), AbstractRunner)
    )


def running_parallel():
    from kedro.runner import ParallelRunner

    return isinstance(get_kedro_runner(), ParallelRunner)


datasets_dict = {}
try:
    from kedro.extras.datasets.json import JSONDataSet

    datasets_dict["json"] = JSONDataSet
except ImportError:
    pass
try:
    from kedro.extras.datasets.pandas import CSVDataSet, ExcelDataSet, ParquetDataSet

    datasets_dict["csv"] = CSVDataSet
    datasets_dict["xls"] = ExcelDataSet
    datasets_dict["parquet"] = ParquetDataSet
except ImportError:
    pass
try:
    from kedro.extras.datasets.pickle import PickleDataSet

    datasets_dict["pkl"] = PickleDataSet
    datasets_dict["pickle"] = PickleDataSet

except ImportError:
    pass
try:
    from kedro.extras.datasets.pillow import ImageDataSet

    datasets_dict["png"] = ImageDataSet
    datasets_dict["jpg"] = ImageDataSet
    datasets_dict["jpeg"] = ImageDataSet
    datasets_dict["img"] = ImageDataSet
except ImportError:
    pass
try:
    from kedro.extras.datasets.text import TextDataSet

    datasets_dict["txt"] = TextDataSet
except ImportError:
    pass
try:
    from kedro.extras.datasets.yaml import YAMLDataSet

    datasets_dict["yml"] = YAMLDataSet
    datasets_dict["yaml"] = YAMLDataSet
except ImportError:
    pass


def mlflow_log_dataset(dataset, enable_mlflow=True):
    fp = getattr(dataset, "_filepath", None)
    if not fp:
        low_ds = getattr(dataset, "_dataset", None)
        if low_ds:
            fp = getattr(low_ds, "_filepath", None)
    if not fp:
        log.warning("_filepath of '{}' was not found.".format(d))
        return
    mlflow_log_artifacts([fp], enable_mlflow=enable_mlflow)


class MLflowCatalogLoggerHook:
    """Logs datasets to MLflow"""

    _logged_set = set()

    def __init__(
        self,
        auto: bool = True,
        mlflow_catalog: Dict[str, Union[str, AbstractDataSet]] = {},
        enable_mlflow: bool = True,
    ):
        """
        Args:
            auto: If True, each dataset (Python func input/output) not listed in the catalog
            will be logged following the same rule as "a" option below.
            mlflow_catalog: [Deprecated in favor of MLflowDataSet] Specify how to log each dataset
            (Python func input/output).

                - If set to "p", the value will be saved/loaded as an MLflow parameter (string).

                - If set to "m", the value will be saved/loaded as an MLflow metric (numeric).

                - If set to "a", the value will be saved/loaded based on the data type.

                    - If the data type is either {float, int}, the value will be saved/loaded as an MLflow metric.

                    - If the data type is either {str, list, tuple, set}, the value will be saved/load as an MLflow parameter.

                    - If the data type is dict, the value will be flattened with dot (".") as the separator and then saved/loaded as either an MLflow metric or parameter based on each data type as explained above.

                - If set to either {"json", "csv", "xls", "parquet", "png", "jpg", "jpeg", "img", "pkl", "txt", "yml", "yaml"}, the backend dataset instance will be created accordingly to save/load as an MLflow artifact.

                - If set to a Kedro DataSet object or a dictionary, it will be used as the backend dataset to save/load as an MLflow artifact.

                - If set to None (default), MLflow logging will be skipped.
            enable_mlflow: Enable logging to MLflow.
        """
        self.enable_mlflow = find_spec("mlflow") and enable_mlflow
        self.mlflow_catalog = mlflow_catalog
        self.auto = auto

    @hook_impl
    def before_pipeline_run(
        self, run_params: Dict[str, Any], pipeline: Pipeline, catalog: DataCatalog
    ) -> None:
        for dataset_name in catalog._data_sets:
            if catalog._data_sets[dataset_name].__class__.__name__ == "MLflowDataSet":
                setattr(catalog._data_sets[dataset_name], "_dataset_name", dataset_name)
                setattr(
                    catalog._data_sets[dataset_name],
                    "_running_parallel",
                    running_parallel(),
                )
                catalog._data_sets[dataset_name]._init_dataset()

    @hook_impl
    def after_node_run(
        self,
        node: Node,
        catalog: DataCatalog,
        inputs: Dict[str, Any],
        outputs: Dict[str, Any],
    ):

        for name, value in inputs.items():
            if name not in self._logged_set:
                self._logged_set.add(name)
                self._log_dataset(name, value)

        for name, value in outputs.items():
            if name not in self._logged_set:
                self._logged_set.add(name)
                self._log_dataset(name, value)

    def _log_dataset(self, name: str, value: Any):
        if name not in self.mlflow_catalog:
            if not self.auto:
                return

            mlflow_log_values({name: value}, enable_mlflow=self.enable_mlflow)
            return

        catalog_instance = self.mlflow_catalog.get(name, None)
        if not catalog_instance:
            return
        elif isinstance(catalog_instance, str):
            if catalog_instance in {"p"}:
                mlflow_log_params({name: value}, enable_mlflow=self.enable_mlflow)
            elif catalog_instance in {"m"}:
                mlflow_log_metrics({name: value}, enable_mlflow=self.enable_mlflow)
            elif catalog_instance in {"a"}:
                mlflow_log_values({name: value}, enable_mlflow=self.enable_mlflow)
            elif catalog_instance in datasets_dict:
                ds = datasets_dict.get(catalog_instance)
                fp = tempfile.gettempdir() + "/" + name + "." + catalog_instance
                ds(filepath=fp).save(value)
                mlflow_log_artifacts([fp], enable_mlflow=self.enable_mlflow)
            else:
                log.warning(
                    "'{}' is not supported as mlflow_catalog entry and ignored.".format(
                        catalog_instance
                    )
                )
                return
        else:
            if not hasattr(catalog_instance, "save"):
                log.warning("'save' attr is not found in mlflow_catalog instance.")
            catalog_instance.save(value)
            mlflow_log_dataset(catalog_instance, enable_mlflow=self.enable_mlflow)



from importlib.util import find_spec
from logging import getLogger
import time
from typing import Any, Callable  # NOQA

from kedro.io import AbstractTransformer

log = getLogger(__name__)


class MLflowIOTimeLoggerTransformer(AbstractTransformer):
    """Log duration time to load and save each dataset."""

    def __init__(
        self,
        enable_mlflow: bool = True,
        metric_name_prefix: str = "_time_to_",
    ):
        """
        Args:
            enable_mlflow: Enable logging to MLflow.
            metric_name_prefix: Prefix for the metric names. The metric names are
                `metric_name_prefix` concatenated with 'load <data_set_name>' or
                'save <data_set_name>'
        """
        self.enable_mlflow = find_spec("mlflow") and enable_mlflow
        self.metric_name_prefix = metric_name_prefix

    def _log_time(self, time_begin, data_set_name, action):
        time_ = time.time() - time_begin
        data_set_name = data_set_name.replace(":", "..")
        time_dict = {
            "{}{} {}".format(self.metric_name_prefix, action, data_set_name)[
                :250
            ]: time_
        }

        log.info("Time duration: {}".format(time_dict))
        if self.enable_mlflow:
            from mlflow import log_metrics

            log_metrics(time_dict)

    def load(self, data_set_name: str, load: Callable[[], Any]) -> Any:
        time_begin = time.time()
        data = load()
        self._log_time(time_begin, data_set_name, "load")
        return data

    def save(self, data_set_name: str, save: Callable[[Any], None], data: Any) -> None:
        time_begin = time.time()
        save(data)
        self._log_time(time_begin, data_set_name, "save")



from datetime import datetime, timedelta
import os
import tempfile

import torch

import logging

log = logging.getLogger(__name__)

__all__ = ["FlexibleModelCheckpoint"]

"""
Copied from https://github.com/pytorch/ignite/blob/v0.2.1/ignite/handlers/checkpoint.py
due to the change in ignite v0.3.0
"""


class ModelCheckpoint(object):
    """ModelCheckpoint handler can be used to periodically save objects to disk.
    This handler expects two arguments:
        - an :class:`~ignite.engine.Engine` object
        - a `dict` mapping names (`str`) to objects that should be saved to disk.
    See Notes and Examples for further details.
    Args:
        dirname (str):
            Directory path where objects will be saved.
        filename_prefix (str):
            Prefix for the filenames to which objects will be saved. See Notes
            for more details.
        save_interval (int, optional):
            if not None, objects will be saved to disk every `save_interval` calls to the handler.
            Exactly one of (`save_interval`, `score_function`) arguments must be provided.
        score_function (callable, optional):
            if not None, it should be a function taking a single argument,
            an :class:`~ignite.engine.Engine` object,
            and return a score (`float`). Objects with highest scores will be retained.
            Exactly one of (`save_interval`, `score_function`) arguments must be provided.
        score_name (str, optional):
            if `score_function` not None, it is possible to store its absolute value using `score_name`. See Notes for
            more details.
        n_saved (int, optional):
            Number of objects that should be kept on disk. Older files will be removed.
        atomic (bool, optional):
            If True, objects are serialized to a temporary file,
            and then moved to final destination, so that files are
            guaranteed to not be damaged (for example if exception occures during saving).
        require_empty (bool, optional):
            If True, will raise exception if there are any files starting with `filename_prefix`
            in the directory 'dirname'.
        create_dir (bool, optional):
            If True, will create directory 'dirname' if it doesnt exist.
        save_as_state_dict (bool, optional):
            If True, will save only the `state_dict` of the objects specified, otherwise the whole object will be saved.
    Note:
          This handler expects two arguments: an :class:`~ignite.engine.Engine` object and a `dict`
          mapping names to objects that should be saved.
          These names are used to specify filenames for saved objects.
          Each filename has the following structure:
          `{filename_prefix}_{name}_{step_number}.pth`.
          Here, `filename_prefix` is the argument passed to the constructor,
          `name` is the key in the aforementioned `dict`, and `step_number`
          is incremented by `1` with every call to the handler.
          If `score_function` is provided, user can store its absolute value using `score_name` in the filename.
          Each filename can have the following structure:
          `{filename_prefix}_{name}_{step_number}_{score_name}={abs(score_function_result)}.pth`.
          For example, `score_name="val_loss"` and `score_function` that returns `-loss` (as objects with highest scores
          will be retained), then saved models filenames will be `model_resnet_10_val_loss=0.1234.pth`.
    Examples:
        >>> import os
        >>> from ignite.engine import Engine, Events
        >>> from ignite.handlers import ModelCheckpoint
        >>> from torch import nn
        >>> trainer = Engine(lambda batch: None)
        >>> handler = ModelCheckpoint('/tmp/models', 'myprefix', save_interval=2, n_saved=2, create_dir=True)
        >>> model = nn.Linear(3, 3)
        >>> trainer.add_event_handler(Events.EPOCH_COMPLETED, handler, {'mymodel': model})
        >>> trainer.run([0], max_epochs=6)
        >>> os.listdir('/tmp/models')
        ['myprefix_mymodel_4.pth', 'myprefix_mymodel_6.pth']
    """

    def __init__(
        self,
        dirname,
        filename_prefix,
        save_interval=None,
        score_function=None,
        score_name=None,
        n_saved=1,
        atomic=True,
        require_empty=True,
        create_dir=True,
        save_as_state_dict=True,
    ):

        self._dirname = os.path.expanduser(dirname)
        self._fname_prefix = filename_prefix
        self._n_saved = n_saved
        self._save_interval = save_interval
        self._score_function = score_function
        self._score_name = score_name
        self._atomic = atomic
        self._saved = []  # list of tuples (priority, saved_objects)
        self._iteration = 0
        self._save_as_state_dict = save_as_state_dict

        if not (save_interval is None) ^ (score_function is None):
            raise ValueError(
                "Exactly one of `save_interval`, or `score_function` "
                "arguments must be provided."
            )

        if score_function is None and score_name is not None:
            raise ValueError(
                "If `score_name` is provided, then `score_function` "
                "should be also provided."
            )

        if create_dir:
            if not os.path.exists(dirname):
                os.makedirs(dirname)

        # Ensure that dirname exists
        if not os.path.exists(dirname):
            raise ValueError("Directory path '{}' is not found.".format(dirname))

        if require_empty:
            matched = [
                fname
                for fname in os.listdir(dirname)
                if fname.startswith(self._fname_prefix)
            ]

            if len(matched) > 0:
                raise ValueError(
                    "Files prefixed with {} are already present "
                    "in the directory {}. If you want to use this "
                    "directory anyway, pass `require_empty=False`."
                    "".format(filename_prefix, dirname)
                )

    def _save(self, obj, path):
        if not self._atomic:
            self._internal_save(obj, path)
        else:
            tmp = tempfile.NamedTemporaryFile(delete=False, dir=self._dirname)
            try:
                self._internal_save(obj, tmp.file)
            except BaseException:
                tmp.close()
                os.remove(tmp.name)
                raise
            else:
                tmp.close()
                os.rename(tmp.name, path)

    def _internal_save(self, obj, path):
        if not self._save_as_state_dict:
            torch.save(obj, path)
        else:
            if not hasattr(obj, "state_dict") or not callable(obj.state_dict):
                raise ValueError("Object should have `state_dict` method.")
            torch.save(obj.state_dict(), path)

    def __call__(self, engine, to_save):
        if len(to_save) == 0:
            raise RuntimeError("No objects to checkpoint found.")

        self._iteration += 1

        if self._score_function is not None:
            priority = self._score_function(engine)

        else:
            priority = self._iteration
            if (self._iteration % self._save_interval) != 0:
                return

        if (len(self._saved) < self._n_saved) or (self._saved[0][0] < priority):
            saved_objs = []

            suffix = ""
            if self._score_name is not None:
                suffix = "_{}={:.7}".format(self._score_name, abs(priority))

            for name, obj in to_save.items():
                fname = "{}_{}_{}{}.pth".format(
                    self._fname_prefix, name, self._iteration, suffix
                )
                path = os.path.join(self._dirname, fname)

                self._save(obj=obj, path=path)
                saved_objs.append(path)

            self._saved.append((priority, saved_objs))
            self._saved.sort(key=lambda item: item[0])

        if len(self._saved) > self._n_saved:
            _, paths = self._saved.pop(0)
            for p in paths:
                os.remove(p)


class FlexibleModelCheckpoint(ModelCheckpoint):
    def __init__(
        self,
        dirname,
        filename_prefix,
        offset_hours=0,
        filename_format=None,
        suffix_format=None,
        *args,
        **kwargs
    ):
        if "%" in filename_prefix:
            filename_prefix = get_timestamp(
                fmt=filename_prefix, offset_hours=offset_hours
            )
        super().__init__(dirname, filename_prefix, *args, **kwargs)
        if not callable(filename_format):
            if isinstance(filename_format, str):
                format_str = filename_format
            else:
                format_str = "{}_{}_{:06d}{}.pth"

            def filename_format(filename_prefix, name, step_number, suffix):
                return format_str.format(filename_prefix, name, step_number, suffix)

        self._filename_format = filename_format

        if not callable(suffix_format):
            if isinstance(suffix_format, str):
                suffix_str = suffix_format
            else:
                suffix_str = "_{}_{:.7}"

            def suffix_format(score_name, abs_priority):
                return suffix_str.format(score_name, abs_priority)

        self._suffix_format = suffix_format

    def __call__(self, engine, to_save):
        if len(to_save) == 0:
            raise RuntimeError("No objects to checkpoint found.")

        self._iteration += 1

        if self._score_function is not None:
            priority = self._score_function(engine)

        else:
            priority = self._iteration
            if (self._iteration % self._save_interval) != 0:
                return

        if (len(self._saved) < self._n_saved) or (self._saved[0][0] < priority):
            saved_objs = []

            suffix = ""
            if self._score_name is not None:
                suffix = self._suffix_format(self._score_name, abs(priority))

            for name, obj in to_save.items():
                fname = self._filename_format(
                    self._fname_prefix, name, self._iteration, suffix
                )
                path = os.path.join(self._dirname, fname)

                self._save(obj=obj, path=path)
                saved_objs.append(path)

            self._saved.append((priority, saved_objs))
            self._saved.sort(key=lambda item: item[0])

        if len(self._saved) > self._n_saved:
            _, paths = self._saved.pop(0)
            for p in paths:
                os.remove(p)


def get_timestamp(fmt="%Y-%m-%dT%H:%M:%S", offset_hours=0):
    return (datetime.now() + timedelta(hours=offset_hours)).strftime(fmt)



import copy
from pathlib import Path
from typing import Any, Dict
import logging

import torchvision

from pipelinex.hatch_dict.hatch_dict import HatchDict
from ..core import AbstractVersionedDataSet, DataSetError, Version

log = logging.getLogger(__name__)


class IterableImagesDataSet(AbstractVersionedDataSet):
    """Loads a folder containing images as an iterable.
    Wrapper of:
    https://pytorch.org/docs/stable/torchvision/datasets.html#imagefolder
    """

    def __init__(
        self,
        filepath: str,
        load_args: Dict[str, Any] = None,
        save_args: Dict[str, Any] = None,
        version: Version = None,
    ) -> None:
        """
        Args:
            filepath: `root` fed to:
                https://pytorch.org/docs/stable/torchvision/datasets.html#imagefolder
            load_args: Args fed to:
                https://pytorch.org/docs/stable/torchvision/datasets.html#imagefolder
            save_args: Ignored as saving is not supported.
            version: If specified, should be an instance of
                ``kedro.io.core.Version``. If its ``load`` attribute is
                None, the latest version will be loaded.
        """

        super().__init__(
            filepath=Path(filepath), version=version, exists_function=self._exists
        )
        self._load_args = load_args
        self._save_args = save_args

    def _load(self) -> Any:

        load_path = Path(self._get_load_path())

        load_args = copy.deepcopy(self._load_args)
        load_args = load_args or dict()

        load_args = HatchDict(load_args).get()

        load_args.setdefault("root", load_path)
        load_args.setdefault("transform", torchvision.transforms.ToTensor())

        vision_dataset = torchvision.datasets.ImageFolder(**load_args)

        return vision_dataset

    def _save(self, vision_dataset) -> None:
        """ Not Implemented """
        return None

    def _describe(self) -> Dict[str, Any]:
        return dict(
            filepath=self._filepath,
            load_args=self._save_args,
            save_args=self._save_args,
            version=self._version,
        )

    def _exists(self) -> bool:
        try:
            path = self._get_load_path()
        except DataSetError:
            return False
        return Path(path).exists()



from logging import getLogger


log = getLogger(__name__)

try:
    from kedro.framework.context import *  # NOQA

    from kedro.framework.hooks import get_hook_manager

except (ImportError, ModuleNotFoundError):
    log.warning(
        "Failed to import context and hooks from kedro.framework."
        "Hooks will not be registered. Recommended to update Kedro to >= 0.16"
    )
    # if kedro.__version__ <= "0.15.9"

    from kedro.context import *  # NOQA

    def get_hook_manager():
        class Hook:
            def after_catalog_created(self, *args, **kwargs):
                pass

            def before_node_run(self, *args, **kwargs):
                pass

            def after_node_run(self, *args, **kwargs):
                pass

            def on_node_error(self, *args, **kwargs):
                pass

            def before_pipeline_run(self, *args, **kwargs):
                pass

            def after_pipeline_run(self, *args, **kwargs):
                pass

            def on_pipeline_error(self, *args, **kwargs):
                pass

        hook = Hook()

        class HookManager:
            def __init__(self):
                self.hook = hook

            def is_registered(self, *args, **kwargs):
                return False

            def register(self, *args, **kwargs):
                pass

        return HookManager()



from importlib.util import find_spec
from logging import getLogger
from typing import Any, Dict, List  # NOQA

from kedro.io import DataCatalog
from kedro.pipeline import Pipeline
from kedro.pipeline.node import Node

from .mlflow_utils import hook_impl, mlflow_log_artifacts


log = getLogger(__name__)


class MLflowArtifactsLoggerHook:
    """Logs artifacts of specified file paths and dataset names to MLflow"""

    def __init__(
        self,
        filepaths_before_pipeline_run: List[str] = None,
        filepaths_after_pipeline_run: List[str] = None,
        datasets_after_node_run: List[str] = None,
        enable_mlflow: bool = True,
    ):
        """
        Args:
            filepaths_before_pipeline_run: The file paths of artifacts to log
                before the pipeline is run.
            filepaths_after_pipeline_run: The file paths of artifacts to log
                after the pipeline is run.
            datasets_after_node_run: The dataset names to log after the node is run.
            enable_mlflow: Enable logging to MLflow.
        """
        self.enable_mlflow = find_spec("mlflow") and enable_mlflow
        self.filepaths_before_pipeline_run = filepaths_before_pipeline_run or []
        self.datasets_after_node_run = datasets_after_node_run or []
        self.filepaths_after_pipeline_run = filepaths_after_pipeline_run or []

    def _log_artifacts(self, artifacts=[]):
        mlflow_log_artifacts(artifacts, enable_mlflow=self.enable_mlflow)

    @hook_impl
    def before_pipeline_run(
        self, run_params: Dict[str, Any], pipeline: Pipeline, catalog: DataCatalog
    ):
        self._log_artifacts(self.filepaths_before_pipeline_run)

    @hook_impl
    def after_pipeline_run(
        self, run_params: Dict[str, Any], pipeline: Pipeline, catalog: DataCatalog
    ):
        self._log_artifacts(self.filepaths_after_pipeline_run)

    def _log_datasets(self, catalog, datasets):
        for d in datasets:
            ds = getattr(catalog.datasets, d, None)
            if not ds:
                log.warning("{} was not found in catalog.datasets.".format(d))
                return
            fp = getattr(ds, "_filepath", None)
            if not fp:
                low_ds = getattr(ds, "_dataset", None)
                if low_ds:
                    fp = getattr(low_ds, "_filepath", None)
            if not fp:
                log.warning("_filepath of '{}' was not found.".format(d))
                return
            mlflow_log_artifacts(fp, enable_mlflow=self.enable_mlflow)

    @hook_impl
    def after_node_run(
        self,
        node: Node,
        catalog: DataCatalog,
        inputs: Dict[str, Any],
        outputs: Dict[str, Any],
    ):
        datasets = [d for d in outputs.keys() if d in self.datasets_after_node_run]
        self._log_datasets(catalog, datasets)



from typing import Any, Dict
import pandas as pd
import matplotlib.pyplot as plt
import logging

from ..core import AbstractDataSet

log = logging.getLogger(__name__)


class HistgramDataSet(AbstractDataSet):
    def __init__(
        self,
        filepath: str,
        save_args: Dict[str, Any] = None,
        hist_args: Dict[str, Any] = None,
    ) -> None:
        self._filepath = filepath
        self._save_args = save_args

    def _describe(self) -> Dict[str, Any]:
        return dict(
            filepath=self._filepath,
            save_args=self._save_args,
        )

    def _load(self) -> Any:
        """ loading is not supported. """
        return None

    def _save(self, df: pd.DataFrame) -> None:
        save_args = self._save_args
        savefig_args = save_args.pop("savefig_args", {})
        df.hist(**save_args)
        plt.savefig(fname=self._filepath, **savefig_args)



import numpy as np

_to_channel_last_dict = {3: (-2, -1, -3), 4: (0, -2, -1, -3)}


def to_channel_last_arr(a):
    if a.ndim in {3, 4}:
        return np.transpose(a, axes=_to_channel_last_dict.get(a.ndim))
    else:
        return a


_to_channel_first_dict = {3: (-1, -3, -2), 4: (0, -1, -3, -2)}


def to_channel_first_arr(a):
    if a.ndim in {3, 4}:
        return np.transpose(a, axes=_to_channel_first_dict.get(a.ndim))
    else:
        return a


def reverse_channel(a, channel_first=False):
    if a.ndim == 3:
        if channel_first:
            return a[::-1, :, :]
        else:
            return a[:, :, ::-1]
    if a.ndim == 4:
        if channel_first:
            return a[:, ::-1, :, :]
        else:
            return a[:, :, :, ::-1]
    return a


class ReverseChannel:
    def __init__(self, channel_first=False):
        self.channel_first = channel_first

    def __call__(self, a):
        return reverse_channel(a, channel_first=self.channel_first)



import sklearn
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import QuantileTransformer, StandardScaler, MinMaxScaler
import pandas as pd
import numpy as np


sklearn_version = getattr(sklearn, "__version__")

if sklearn_version and sklearn_version < "0.22":

    class EstimatorTransformer(BaseEstimator, TransformerMixin):
        pass


else:

    class EstimatorTransformer(TransformerMixin, BaseEstimator):
        pass


def extract_from_df(df, cols, target_col):
    if isinstance(df, pd.DataFrame):
        cols = cols or df.columns
        target_col = target_col
        X = df[cols].values
        y = df[target_col].values if target_col else None
    else:
        X = df
        y = None
    if y:
        return dict(X=X, y=y)
    else:
        return dict(X=X), cols


class ZeroToZeroTransformer(EstimatorTransformer):
    def __init__(self, zero_to_zero=False, **kwargs):
        self.zero_to_zero = zero_to_zero
        super().__init__(**kwargs)

    def _keep_zero(self, X):
        zeros_arr = np.zeros((1, X.shape[1]))
        offset_arr = super().transform(zeros_arr)
        X = X - offset_arr
        return X

    def transform(self, X):
        out_X = super().transform(X)
        if self.zero_to_zero:
            out_X = self._keep_zero(out_X)
        return out_X

    def fit_transform(self, X):
        out_X = super().fit_transform(X)
        if self.zero_to_zero:
            out_X = self._keep_zero(out_X)
        return out_X


class DfBaseTransformer(ZeroToZeroTransformer):
    def __init__(self, cols=None, target_col=None, **kwargs):
        super().__init__(**kwargs)
        self.cols = cols
        self.target_col = target_col

    def _assign_to_df(self, df, cols, out_arr):
        if isinstance(df, pd.DataFrame):
            if getattr(self, "copy", True):
                out_df = df.copy()
            else:
                out_df = df
            out_df.loc[:, cols] = out_arr
            return out_df
        else:
            return out_arr

    def fit(self, df):
        d, cols = extract_from_df(df, self.cols, self.target_col)
        return super().fit(**d)

    def transform(self, df):
        d, cols = extract_from_df(df, self.cols, self.target_col)
        out_arr = super().transform(**d)
        return self._assign_to_df(df, cols, out_arr)

    def fit_transform(self, df):
        d, cols = extract_from_df(df, self.cols, self.target_col)
        out_arr = super().fit_transform(**d)
        return self._assign_to_df(df, cols, out_arr)

    def __call__(self, df):
        return self.fit_transform(df), self


class DfQuantileTransformer(DfBaseTransformer, QuantileTransformer):
    pass


class DfStandardScaler(DfBaseTransformer, StandardScaler):
    pass


class DfMinMaxScaler(DfBaseTransformer, MinMaxScaler):
    pass


class DfTrainTestSplit:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, df, *argsignore, **kwargsignore):
        return sklearn.model_selection.train_test_split(df, **self.kwargs)



import asyncio
import sys
from typing import Any, Dict

import httpx

from ..requests.api_dataset import APIDataSet


def asyncio_run(aw):
    if sys.version_info >= (3, 7):
        return asyncio.run(aw)

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        r = loop.run_until_complete(aw)
    finally:
        loop.close()
        asyncio.set_event_loop(None)

    return r


async def request_coroutine(session, method, url, request_args):
    return await session.request(method, url=url, **request_args)


async def requests_coroutine(session_config, method, url_list, request_args):
    async with httpx.AsyncClient(**session_config) as session:
        request_coroutines = [
            request_coroutine(session, method, url, request_args) for url in url_list
        ]

        request_tasks = [
            asyncio.ensure_future(coroutine) for coroutine in request_coroutines
        ]
        r = await asyncio.wait(request_tasks)
    return r


class AsyncAPIDataSet(APIDataSet):
    def _configure_session(self, session_config, _):
        return session_config

    def _execute_request(self) -> Dict[str, Any]:

        request_args = self._request_args
        method = self._method
        url_dict = self._get_url_dict()
        session_config = self._session

        name_url_list = list(url_dict.items())
        url_list = [e[1] for e in name_url_list]
        tasks_done, tasks_pending = asyncio_run(
            requests_coroutine(session_config, method, url_list, request_args)
        )

        name_list = [e[0] for e in name_url_list]
        response_dict = {}
        for name, task in zip(name_list, tasks_done):
            try:
                response_dict[name] = task.result()
            except Exception as exc:
                response_dict[name] = self._handle_exceptions(exc)

        return response_dict



import logging
from importlib.util import find_spec
from pathlib import Path
import tempfile
from typing import Any, Dict, Union

from kedro.io.core import AbstractDataSet
from kedro.io import MemoryDataSet
from kedro.extras.datasets.pickle import PickleDataSet

from pipelinex.mlflow_on_kedro.hooks.mlflow.mlflow_utils import (
    mlflow_log_artifacts,
    mlflow_log_metrics,
    mlflow_log_params,
    mlflow_log_values,
)

log = logging.getLogger(__name__)

dataset_dicts = {
    "json": {"type": "json.JSONDataSet"},
    "csv": {"type": "pandas.CSVDataSet"},
    "xls": {"type": "pandas.ExcelDataSet"},
    "parquet": {"type": "pandas.ParquetDataSet"},
    "pkl": {"type": "pickle.PickleDataSet"},
    "png": {"type": "pillow.ImageDataSet"},
    "jpg": {"type": "pillow.ImageDataSet"},
    "jpeg": {"type": "pillow.ImageDataSet"},
    "img": {"type": "pillow.ImageDataSet"},
    "txt": {"type": "text.TextDataSet"},
    "yaml": {"type": "yaml.YAMLDataSet"},
    "yml": {"type": "yaml.YAMLDataSet"},
}


class MLflowDataSet(AbstractDataSet):
    """``MLflowDataSet`` saves data to, and loads data from MLflow.

    You can also specify a ``MLflowDataSet`` in catalog.yml

    Example:
    ::

        >>> test_ds:
        >>>    type: MLflowDataSet
        >>>    dataset: pkl
    """

    def __init__(
        self,
        dataset: Union[AbstractDataSet, Dict, str] = None,
        filepath: str = None,
        dataset_name: str = None,
        saving_tracking_uri: str = None,
        saving_experiment_name: str = None,
        saving_run_id: str = None,
        loading_tracking_uri: str = None,
        loading_run_id: str = None,
        caching: bool = True,
        copy_mode: str = None,
        file_caching: bool = True,
    ):
        """

        Args:
            dataset: Specify how to treat the dataset as an MLflow metric, parameter, or artifact.

                - If set to "p", the value will be saved/loaded as an MLflow parameter (string).

                - If set to "m", the value will be saved/loaded as an MLflow metric (numeric).

                - If set to "a", the value will be saved/loaded based on the data type.

                    - If the data type is either {float, int}, the value will be saved/loaded as an MLflow metric.

                    - If the data type is either {str, list, tuple, set}, the value will be saved/load as an MLflow parameter.

                    - If the data type is dict, the value will be flattened with dot (".") as the separator and then saved/loaded as either an MLflow metric or parameter based on each data type as explained above.

                - If set to either {"json", "csv", "xls", "parquet", "png", "jpg", "jpeg", "img", "pkl", "txt", "yml", "yaml"}, the backend dataset instance will be created accordingly to save/load as an MLflow artifact.

                - If set to a Kedro DataSet object or a dictionary, it will be used as the backend dataset to save/load as an MLflow artifact.

                - If set to None (default), MLflow logging will be skipped.
            filepath: File path, usually in local file system, to save to and load from.
                Used only if the dataset arg is a string.
                If None (default), ``<temp directory>/<dataset_name arg>.<dataset arg>`` is used.
            dataset_name: Used only if the dataset arg is a string and filepath arg is None.
                If None (default), Python object ID is used, but will be overwritten by
                MLflowCatalogLoggerHook.
            saving_tracking_uri: MLflow Tracking URI to save to.
                If None (default), MLFLOW_TRACKING_URI environment variable is used.
            saving_experiment_name: MLflow experiment name to save to.
                If None (default), new experiment will not be created or started.
                Ignored if saving_run_id is set.
            saving_run_id: An existing MLflow experiment run ID to save to.
                If None (default), no existing experiment run will be resumed.
            loading_tracking_uri: MLflow Tracking URI to load from.
                If None (default), MLFLOW_TRACKING_URI environment variable is used.
            loading_run_id: MLflow experiment run ID to load from.
                If None (default), current active run ID will be used if available.
            caching: Enable caching if parallel runner is not used. True in default.
            copy_mode: The copy mode used to copy the data. Possible
                values are: "deepcopy", "copy" and "assign". If not
                provided, it is inferred based on the data type.
                Ignored if caching arg is False.
            file_caching: Attempt to use the file at filepath when loading if no cache found
                in memory. True in default.

        """
        self.dataset = dataset or MemoryDataSet()
        self.filepath = filepath
        self.dataset_name = dataset_name
        self.saving_tracking_uri = saving_tracking_uri
        self.saving_experiment_name = saving_experiment_name
        self.saving_run_id = saving_run_id
        self.loading_tracking_uri = loading_tracking_uri
        self.loading_run_id = loading_run_id
        self.caching = caching
        self.file_caching = file_caching
        self.copy_mode = copy_mode
        self._dataset_name = str(id(self))

        if isinstance(dataset, str):
            if (dataset not in {"p", "m"}) and (dataset not in dataset_dicts):
                raise ValueError(
                    "`dataset`: {} not supported. Specify one of {}.".format(
                        dataset, list(dataset_dicts.keys())
                    )
                )
        self._ready = False
        self._running_parallel = None
        self._cache = None

    def _init_dataset(self):

        if not getattr(self, "_ready", None):
            self._ready = True
            self.dataset_name = self.dataset_name or self._dataset_name
            _dataset = self.dataset
            if isinstance(self.dataset, str):
                dataset_dict = dataset_dicts.get(
                    self.dataset, {"type": "pickle.PickleDataSet"}
                )
                dataset_dict["filepath"] = self.filepath = (
                    self.filepath
                    or tempfile.gettempdir()
                    + "/"
                    + self.dataset_name
                    + "."
                    + self.dataset
                )
                _dataset = dataset_dict

            if isinstance(_dataset, dict):
                self._dataset = AbstractDataSet.from_config(
                    self._dataset_name, _dataset
                )
            elif isinstance(_dataset, AbstractDataSet):
                self._dataset = _dataset
            else:
                raise ValueError(
                    "The argument type of `dataset` should be either a dict/YAML "
                    "representation of the dataset, or the actual dataset object."
                )

            _filepath = getattr(self._dataset, "_filepath", None)
            if _filepath:
                self.filepath = str(_filepath)

            if self.caching and (not self._running_parallel):
                self._cache = MemoryDataSet(copy_mode=self.copy_mode)

    def _release(self) -> None:
        self._init_dataset()
        self._dataset.release()
        if self._cache:
            self._cache.release()

    def _describe(self) -> Dict[str, Any]:
        return {
            "dataset": self._dataset._describe()
            if getattr(self, "_ready", None)
            else self.dataset,  # pylint: disable=protected-access
            "filepath": self.filepath,
            "saving_tracking_uri": self.saving_tracking_uri,
            "saving_experiment_name": self.saving_experiment_name,
            "saving_run_id": self.saving_run_id,
            "loading_tracking_uri": self.loading_tracking_uri,
            "loading_run_id": self.loading_run_id,
        }

    def _load(self):
        self._init_dataset()

        if self._cache and self._cache.exists():
            return self._cache.load()

        if self.file_caching and self._dataset.exists():
            return self._dataset.load()

        import mlflow

        client = mlflow.tracking.MlflowClient(tracking_uri=self.loading_tracking_uri)

        self.loading_run_id = self.loading_run_id or mlflow.active_run().info.run_id

        if self.dataset in {"p"}:
            run = client.get_run(self.loading_run_id)
            value = run.data.params.get(self.dataset_name, None)
            if value is None:
                raise KeyError(
                    "param '{}' not found in run_id '{}'.".format(
                        self.dataset_name, self.loading_run_id
                    )
                )

            PickleDataSet(filepath=self.filepath).save(value)

        elif self.dataset in {"m"}:
            run = client.get_run(self.loading_run_id)
            value = run.data.metrics.get(self.dataset_name, None)
            if value is None:
                raise KeyError(
                    "metric '{}' not found in run_id '{}'.".format(
                        self.dataset_name, self.loading_run_id
                    )
                )
            PickleDataSet(filepath=self.filepath).save(value)

        else:
            p = Path(self.filepath)

            dst_path = tempfile.gettempdir()
            downloaded_path = client.download_artifacts(
                run_id=self.loading_run_id,
                path=p.name,
                dst_path=dst_path,
            )
            if Path(downloaded_path) != p:
                Path(downloaded_path).rename(p)

        return self._dataset.load()

    def _save(self, data: Any) -> None:
        self._init_dataset()

        self._dataset.save(data)

        if self._cache:
            self._cache.save(data)

        if find_spec("mlflow"):
            import mlflow

            if self.saving_tracking_uri:
                mlflow.set_tracking_uri(self.saving_tracking_uri)

            if self.saving_run_id:
                mlflow.start_run(run_id=self.saving_run_id)

            elif self.saving_experiment_name:
                experiment_id = mlflow.get_experiment_by_name(
                    self.saving_experiment_name
                ).experiment_id
                mlflow.start_run(run_id=self.saving_run_id, experiment_id=experiment_id)

            if self.dataset in {"p"}:
                mlflow_log_params({self.dataset_name: data})
            elif self.dataset in {"m"}:
                mlflow_log_metrics({self.dataset_name: data})
            elif self.dataset in {"a"}:
                mlflow_log_values({self.dataset_name: data})
            else:
                mlflow_log_artifacts([self.filepath])

            if self.saving_run_id or self.saving_experiment_name:
                mlflow.end_run()

    def _exists(self) -> bool:
        self._init_dataset()
        if self._cache:
            return self._cache.exists()
        else:
            return False

    def __getstate__(self):
        return self.__dict__



from pathlib import Path
from typing import Any, Dict
import pandas as pd
from ..core import AbstractVersionedDataSet, DataSetError, Version


class PandasProfilingDataSet(AbstractVersionedDataSet):
    """``PandasProfilingDataSet`` is an ``AbstractVersionedDataSet`` to generate
    pandas profiling report.
    See https://github.com/pandas-profiling/pandas-profiling for details.
    """

    DEFAULT_SAVE_ARGS = dict()  # type: Dict[str, Any]

    def __init__(
        self,
        filepath: str,
        save_args: Dict[str, Any] = None,
        sample_args: Dict[str, Any] = None,
        version: Version = None,
    ) -> None:
        """Creates a new instance of ``PandasProfilingDataSet`` pointing to a concrete
        filepath.

        Args:
            filepath: path to a local yaml file.
            save_args: Arguments passed on to ``df.profile_report`` such as title.
                See https://pandas-profiling.github.io/pandas-profiling/docs/ for details.
                See https://github.com/pandas-profiling/pandas-profiling/blob/master/pandas_profiling/config_default.yaml for default values.
            sample_args: Arguments passed on to ``df.sample``.
                See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.sample.html for details.
            version: If specified, should be an instance of
                ``kedro.io.core.Version``. If its ``load`` attribute is
                None, the latest version will be loaded. If its ``save``
                attribute is None, save version will be autogenerated.

        """
        super().__init__(
            filepath=Path(filepath), version=version, exists_function=self._exists
        )
        self._load_args = {}
        self._save_args = save_args
        self._sample_args = sample_args

    def _describe(self) -> Dict[str, Any]:
        return dict(
            filepath=self._filepath,
            save_args=self._save_args,
            sampling_args=self._sample_args,
            version=self._version,
        )

    def _load(self) -> Any:
        """ loading is not supported. """
        return None

    def _save(self, data: pd.DataFrame) -> None:
        save_path = Path(self._get_save_path())
        save_path.parent.mkdir(parents=True, exist_ok=True)

        if self._sample_args is not None:
            data = data.sample(**self._sample_args)
        profile = data.profile_report(**self._save_args)
        profile.to_file(output_file=save_path)

        load_path = Path(self._get_load_path())
        self._check_paths_consistency(load_path.absolute(), save_path.absolute())

    def _exists(self) -> bool:
        try:
            path = self._get_load_path()
        except DataSetError:
            return False
        return Path(path).is_file()



from importlib.util import find_spec
import os
from logging import getLogger
from typing import List  # NOQA

from .mlflow_utils import hook_impl, mlflow_log_metrics, mlflow_log_params


log = getLogger(__name__)


def env_vars_to_dict(env_vars=[], prefix=""):
    if isinstance(env_vars, str):
        if env_vars == "__ALL__":
            return dict(os.environ)
        else:
            env_vars = [env_vars]
    elif not isinstance(env_vars, (list, tuple, set)):
        raise "{} not supported as env_vars arg".format(type(env_vars))
    return {(prefix + ev): os.environ.get(ev) for ev in env_vars}


def log_param_env_vars(env_vars=[], prefix="", enable_mlflow=True):
    env_var_dict = env_vars_to_dict(env_vars=env_vars, prefix=prefix)
    mlflow_log_params(env_var_dict, enable_mlflow)


def log_metric_env_vars(env_vars=[], prefix="", enable_mlflow=True):
    env_var_dict = env_vars_to_dict(env_vars=env_vars, prefix=prefix)
    metric_env_var_dict = {ev: float(value) for (ev, value) in env_var_dict.items()}
    mlflow_log_metrics(metric_env_var_dict, enable_mlflow)


class MLflowEnvVarsLoggerHook:
    """Logs environment variables to MLflow"""

    def __init__(
        self,
        param_env_vars: List[str] = None,
        metric_env_vars: List[str] = None,
        prefix: str = None,
        enable_mlflow: bool = True,
    ):
        """
        Args:
            param_env_vars: Environment variables to log to MLflow as parameters
            metric_env_vars: Environment variables to log to MLflow as metrics
            prefix: Prefix to add to each name of MLflow parameters and metrics ("env.." in default)
            enable_mlflow: Enable logging to MLflow.
        """
        self.enable_mlflow = find_spec("mlflow") and enable_mlflow
        self.param_env_vars = param_env_vars or []
        self.metric_env_vars = metric_env_vars or []
        self.prefix = prefix or "env.."

    def _log_env_vars(self):
        log_param_env_vars(
            env_vars=self.param_env_vars,
            prefix=self.prefix,
            enable_mlflow=self.enable_mlflow,
        )
        log_metric_env_vars(
            env_vars=self.metric_env_vars,
            prefix=self.prefix,
            enable_mlflow=self.enable_mlflow,
        )

    @hook_impl
    def before_pipeline_run(self):
        self._log_env_vars()

    @hook_impl
    def after_pipeline_run(self):
        self._log_env_vars()



from functools import wraps
from typing import Callable

import kedro

import pipelinex


def task_func1(a):
    return a


def task_func2(a):
    return a


def task_func3(a):
    return a


def stringify(obj):
    print("\n" + obj.__str__())
    print("\n" + obj.to_json())
    s = obj.to_json()
    return s.replace("Flexible", "")


def assert_eq(obj1, obj2):
    assert stringify(obj1) == stringify(obj2)


def task_deco(func: Callable) -> Callable:
    @wraps(func)
    def _task_deco(a):
        a += 10
        result = func(a)
        result += 100
        return result

    return _task_deco


def test_dict():
    fp = pipelinex.FlexiblePipeline(
        nodes=[dict(func=task_func1, inputs="my_input", outputs="my_output")]
    )
    kp = kedro.pipeline.Pipeline(
        nodes=[
            kedro.pipeline.node(func=task_func1, inputs="my_input", outputs="my_output")
        ]
    )
    assert_eq(fp, kp)


def test_sequential():
    fp = pipelinex.FlexiblePipeline(
        nodes=[
            dict(
                func=[task_func1, task_func2, task_func3],
                inputs="my_input",
                outputs="my_output",
            )
        ]
    )
    kp = kedro.pipeline.Pipeline(
        nodes=[
            kedro.pipeline.node(
                func=task_func1, inputs="my_input", outputs="my_output__001"
            ),
            kedro.pipeline.node(
                func=task_func2, inputs="my_output__001", outputs="my_output__002"
            ),
            kedro.pipeline.node(
                func=task_func3, inputs="my_output__002", outputs="my_output"
            ),
        ]
    )
    assert_eq(fp, kp)


def test_decorator():
    if kedro.__version__[:5] not in {"0.16.", "0.17."}:
        return
    fp = pipelinex.FlexiblePipeline(
        nodes=[
            kedro.pipeline.node(func=task_func1, inputs="my_input", outputs="my_output")
        ],
        decorator=[task_deco, task_deco],
    )
    kp = kedro.pipeline.Pipeline(
        nodes=[
            kedro.pipeline.node(func=task_func1, inputs="my_input", outputs="my_output")
        ]
    ).decorate(task_deco, task_deco)
    assert_eq(fp, kp)



from typing import Any, Dict  # NOQA
from importlib import import_module
from logging import getLogger

import kedro
from kedro.pipeline import Pipeline  # NOQA

from pipelinex import HatchDict
from .context import KedroContext, get_hook_manager


log = getLogger(__name__)


class FlexibleParametersContext(KedroContext):
    _params = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        params = self.params
        import_modules(params.pop("IMPORT", []))

        params = HatchDict(params).get()
        hooks = params.pop("HOOKS", None)
        if hooks:
            if kedro.__version__.startswith("0.16."):
                self._hooks_in_params_read = False
                self._hook_manager = getattr(self, "_hook_manager", get_hook_manager())
                self._register_kedro_hooks(hooks)
            else:
                log.warning(
                    "HOOKS defined in config file is ignored as the installed kedro version is not 0.16.x"
                )
        self._kedro_pipelines = params.pop("PIPELINES", None)
        self._kedro_run_config = params.pop("RUN_CONFIG", None) or {}
        self._params = params
        log.info("RUN_CONFIG: \n{}".format(self._kedro_run_config))
        log.debug("PIPELINES: \n{}".format(self._kedro_pipelines))
        log.debug("params: \n{}".format(self._params))

    def _register_kedro_hooks(self, hooks):

        if not self._hooks_in_params_read:
            self._hooks_in_params_read = True

            if not isinstance(hooks, (list, tuple)):
                hooks = [hooks]

            for hook in hooks:
                if not self._hook_manager.is_registered(hook):
                    self._hook_manager.register(hook)
                    log.info(
                        "Registered {} with args: {}".format(
                            getattr(getattr(hook, "__class__", None), "__name__", None),
                            getattr(hook, "__dict__", None),
                        )
                    )

    @property
    def params(self) -> Dict[str, Any]:
        if self._params is None:
            return super().params
        return self._params

    def _get_pipelines(self) -> Dict[str, Pipeline]:
        assert (
            self._kedro_pipelines
        ), "'PIPELINES' key or value is missing in parameters."
        return self._kedro_pipelines

    def run(self, *args, **kwargs):
        run_dict = self._kedro_run_config
        run_dict.update(kwargs)
        log.info("Run pipeline ({})".format(run_dict))
        return super().run(*args, **run_dict)


def import_modules(modules=None):
    if modules:
        if not isinstance(modules, list):
            modules = [modules]
        for module in modules:
            assert isinstance(module, str), "'{}' is not string.".format(module)
            import_module(module)



# Copyright 2020 QuantumBlack Visual Analytics Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
# NONINFRINGEMENT. IN NO EVENT WILL THE LICENSOR OR OTHER CONTRIBUTORS
# BE LIABLE FOR ANY CLAIM, DAMAGES, OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF, OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# The QuantumBlack Visual Analytics Limited ("QuantumBlack") name and logo
# (either separately or in combination, "QuantumBlack Trademarks") are
# trademarks of QuantumBlack. The License does not grant you any right or
# license to the QuantumBlack Trademarks. You may not use the QuantumBlack
# Trademarks or any confusingly similar mark as a trademark for your product,
#     or use the QuantumBlack Trademarks in any other manner that might cause
# confusion in the marketplace, including but not limited to in advertising,
# on websites, or on software.
#
# See the License for the specific language governing permissions and
# limitations under the License.

"""``APIDataSet`` loads the data from HTTP(S) APIs
and returns them into either as string or json Dict.
It uses the python requests library: https://requests.readthedocs.io/en/master/
"""
import socket
from typing import Any, Dict, List, Tuple, Union

import requests
from requests.auth import AuthBase
from ..core import AbstractDataSet, DataSetError


class APIDataSet(AbstractDataSet):
    """``APIDataSet`` loads the data from HTTP(S) APIs.
    It uses the python requests library: https://requests.readthedocs.io/en/master/

    Example:
    ::

        >>> from kedro.extras.datasets.api import APIDataSet
        >>>
        >>>
        >>> data_set = APIDataSet(
        >>>     url="https://quickstats.nass.usda.gov"
        >>>     params={
        >>>         "key": "SOME_TOKEN",
        >>>         "format": "JSON",
        >>>         "commodity_desc": "CORN",
        >>>         "statisticcat_des": "YIELD",
        >>>         "agg_level_desc": "STATE",
        >>>         "year": 2000
        >>>     }
        >>> )
        >>> data = data_set.load()

    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        url: Union[str, List[str], Dict[str, str]] = None,
        method: str = "GET",
        data: Any = None,
        params: Dict[str, Any] = None,
        headers: Dict[str, Any] = None,
        auth: Union[Tuple[str], AuthBase] = None,
        timeout: int = 60,
        attribute: str = "",
        skip_errors: bool = False,
        transforms: List[callable] = [],
        session_config: Dict[str, Any] = {},
        pool_config: Dict[str, Dict[str, Any]] = {
            "https://": {
                "pool_connections": 10,
                "pool_maxsize": 10,
                "max_retries": 0,
                "pool_block": False,
            },
            "http://": {
                "pool_connections": 10,
                "pool_maxsize": 10,
                "max_retries": 0,
                "pool_block": False,
            },
        },
    ) -> None:
        """Creates a new instance of ``APIDataSet`` to fetch data from an API endpoint.

        Args:
            url: The API URL endpoint.
            method: The Method of the request, GET, POST, PUT, DELETE, HEAD, etc...
            data: The request payload, used for POST, PUT, etc requests
                https://requests.readthedocs.io/en/master/user/quickstart/#more-complicated-post-requests
            params: The url parameters of the API.
                https://requests.readthedocs.io/en/master/user/quickstart/#passing-parameters-in-urls
            headers: The HTTP headers.
                https://requests.readthedocs.io/en/master/user/quickstart/#custom-headers
            auth: Anything ``requests`` accepts. Normally it's either ``('login', 'password')``,
                or ``AuthBase``, ``HTTPBasicAuth`` instance for more complex cases.
            timeout: The wait time in seconds for a response, defaults to 1 minute.
                https://requests.readthedocs.io/en/master/user/quickstart/#timeouts
            attribute: The attribute of response to return. Normally it's either `text`, which
                returns pure text,`json`, which returns JSON in Python Dict format, `content`,
                which returns a raw content, or `` (empty string), which returns the response
                object itself. Defaults to `` (empty string).
            skip_errors: If True, exceptions will not interrupt loading data and be returned
                instead of the expected responses by _load method. Defaults to False.
            transforms: List of callables to transform the output.
            session_config: Dict of arguments fed to the session.
            pool_config: Dict of mounting prefix key to Dict of requests.adapters.HTTPAdapter
                param key to value.
                https://requests.readthedocs.io/en/master/user/advanced/#transport-adapters
                https://urllib3.readthedocs.io/en/latest/advanced-usage.html

        """
        super().__init__()
        self._request_args: Dict[str, Any] = {
            "data": data,
            "params": params,
            "headers": headers,
            "auth": auth,
            "timeout": timeout,
        }

        self._url = url
        self._method = method
        self._attribute = attribute
        self._skip_errors = skip_errors
        self._transforms = transforms

        self._session_config = session_config
        self._pool_config = pool_config
        self._session = self._configure_session(session_config, pool_config)

    def _configure_session(self, session_config, pool_config):
        session = requests.Session(**session_config)
        for prefix, adapter_params in pool_config.items():
            session.mount(prefix, requests.adapters.HTTPAdapter(**adapter_params))

        return session

    def _describe(self) -> Dict[str, Any]:
        return dict(
            **self._request_args,
            url=self._url,
            method=self._method,
            session_config=self._session_config,
            pool_config=self._pool_config,
            attribute=self._attribute,
            skip_errors=self._skip_errors,
        )

    def _get_url_dict(self):
        if isinstance(self._url, str):
            url_dict = {"_": self._url}
        elif isinstance(self._url, list):
            url_dict = {i: url for (i, url) in enumerate(self._url)}
        else:
            url_dict = self._url
        return url_dict

    def _execute_request(self) -> Dict[str, requests.Response]:

        request_args = self._request_args
        session = self._session
        method = self._method
        url_dict = self._get_url_dict()

        def request(url):
            response = session.request(method, url=url, **request_args)
            response.raise_for_status()
            return response

        response_dict = {}
        for name, url in url_dict.items():
            try:
                response_dict[name] = request(url)
            except Exception as exc:
                response_dict[name] = self._handle_exceptions(exc)

        return response_dict

    def _handle_exceptions(self, exc):

        if isinstance(exc, requests.exceptions.HTTPError):
            e = DataSetError("Failed to fetch data", exc)
        elif isinstance(exc, socket.error):
            e = DataSetError("Failed to connect to the remote server")
        else:
            e = DataSetError("Exception", exc)

        if self._skip_errors:
            return e
        raise e

    def _load(self) -> Any:
        response_dict = self._execute_request()

        output_dict = {}
        for name, response in response_dict.items():
            if isinstance(response, Exception):
                output_dict[name] = response
                continue
            if response.status_code != requests.codes.ok:
                output_dict[name] = response
                continue

            if not self._attribute:
                output = response
            elif hasattr(response, self._attribute):
                if self._attribute == "json":
                    output = response.json()
                else:
                    output = getattr(response, self._attribute)
            elif self._skip_errors:
                output_dict[name] = response
                continue
            else:
                raise DataSetError(
                    "Response has no attribute: {}".format(self._attribute)
                )

            try:
                for transform in self._transforms:
                    output = transform(output)
                output_dict[name] = output
            except Exception as exc:
                e = DataSetError("Exception", exc)
                if self._skip_errors:
                    output_dict[name] = e
                    continue
                else:
                    raise e

        if isinstance(self._url, str):
            return next(iter(output_dict.values()))
        elif isinstance(self._url, list):
            return [output_dict[i] for i in range(len(output_dict))]
        else:
            return output_dict

    def _save(self, data: Any) -> None:
        raise DataSetError(
            "{} is a read only data set type".format(self.__class__.__name__)
        )

    def _exists(self) -> bool:
        response_dict = self._execute_request()

        return all(
            [
                getattr(response, "status_code") == requests.codes.ok
                for response in response_dict.values()
            ]
        )

    def __call__(
        self,
        url: Union[str, List[str], Dict[str, str]] = None,
        method: str = None,
        data: Any = None,
        params: Dict[str, Any] = None,
        headers: Dict[str, Any] = None,
        auth: Union[Tuple[str], AuthBase] = None,
        timeout: int = None,
        attribute: str = None,
        skip_errors: bool = None,
        transforms: List[callable] = None,
    ):
        if data is not None:
            self._request_args.update({"data": data})
        if params is not None:
            self._request_args.update({"params": params})
        if headers is not None:
            self._request_args.update({"headers": headers})
        if auth is not None:
            self._request_args.update({"auth": auth})
        if timeout is not None:
            self._request_args.update({"timeout": timeout})
        if url is not None:
            self._url = url
        if method is not None:
            self._method = method
        if attribute is not None:
            self._attribute = attribute
        if skip_errors is not None:
            self._skip_errors = skip_errors
        if transforms is not None:
            self._transforms = transforms
        return self._load()



from typing import Dict

from kedro.io import AbstractDataSet, DataCatalog

try:
    from kedro.framework.hooks import hook_impl
except ModuleNotFoundError:

    def hook_impl(func):
        return func


class AddCatalogDictHook:
    """Hook to add data sets."""

    def __init__(
        self,
        catalog_dict: Dict[str, AbstractDataSet],
    ):
        """
        Args:
            catalog_dict: catalog_dict to add.
        """
        assert isinstance(catalog_dict, dict), "{} is not a dict.".format(catalog_dict)
        self._catalog_dict = catalog_dict

    @hook_impl
    def after_catalog_created(self, catalog: DataCatalog) -> None:
        catalog.add_feed_dict(self._catalog_dict)



from typing import List, Tuple, Union

from kedro.io import DataCatalog
from kedro.io.transformers import AbstractTransformer

try:
    from kedro.framework.hooks import hook_impl
except ModuleNotFoundError:

    def hook_impl(func):
        return func


class AddTransformersHook:
    """Hook to add transformers."""

    def __init__(
        self,
        transformers: Union[
            AbstractTransformer, List[AbstractTransformer], Tuple[AbstractTransformer]
        ],
    ):
        """
        Args:
            transformers: transformers to add.
        """
        transformers = transformers or []
        if isinstance(transformers, AbstractTransformer):
            transformers = [transformers]
        self._transformers = transformers

    @hook_impl
    def after_catalog_created(self, catalog: DataCatalog) -> None:
        for t in self._transformers:
            catalog.add_transformer(t)



import re
from codecs import open
from os import path

from setuptools import find_packages, setup

name = "pipelinex"
here = path.abspath(path.dirname(__file__))

with open(path.join(here, "src", name, "__init__.py"), encoding="utf-8") as f:
    result = re.search(r'__version__ = ["\']([^"\']+)', f.read())

    if not result:
        raise ValueError("Can't find the version")

    version = result.group(1)

# get the dependencies and installs
with open("requirements.txt", "r") as f:
    requires = [x.strip() for x in f if x.strip()]

with open("requirements_optional.txt", "r") as f:
    requires_optional = [x.strip() for x in f if x.strip()]
requires_optional += requires

with open("requirements_docs.txt", "r") as f:
    requires_docs = [x.strip() for x in f if x.strip()]
requires_docs += requires_optional

with open("requirements_dev.txt", "r") as f:
    requires_dev = [x.strip() for x in f if x.strip()]
requires_dev += requires_docs

readme = r"""
# PipelineX

GitHub Repository:
https://github.com/Minyus/pipelinex

Documentation:
https://pipelinex.readthedocs.io/
"""

setup(
    name=name,
    version=version,
    description="PipelineX: Python package to build ML pipelines for experimentation with Kedro, MLflow, and more",
    license="Apache Software License (Apache 2.0)",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/Minyus/pipelinex",
    packages=find_packages(where="src", exclude=["tests"]),
    package_dir={"": "src"},
    test_suite="tests",
    install_requires=requires,
    extras_require=dict(
        optional=requires_optional, docs=requires_docs, dev=requires_dev
    ),
    author="Yusuke Minami",
    author_email="me@minyus.github.com",
    zip_safe=False,
    keywords="pipelines, machine learning, deep learning, data science, data engineering",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
)



from importlib.util import find_spec
import json
import time
from logging import getLogger
from pprint import pformat
from pathlib import Path
import tempfile
from typing import Any, Callable, Dict  # NOQA

from kedro.pipeline.node import Node  # NOQA

from .mlflow_utils import hook_impl, mlflow_log_metrics, mlflow_log_artifacts


log = getLogger(__name__)


def _get_task_name(node: Node) -> str:
    func_name = (
        node._func_name.replace("<", "")
        .replace(">", "")
        .split(" ")[0]
        .split(".")[-1][:250]
    )
    return "{} -- {}".format(func_name, " - ".join(node.outputs))


def dump_dict(filepath: str, d: dict):
    with open(filepath, "w") as outfile:
        json.dump(d, outfile)


def load_dict(filepath: str):
    with open(filepath, "r") as outfile:
        d = json.load(outfile)
    return d


class MLflowTimeLoggerHook:
    """
    Logs duration time to run each node (task) to MLflow.
    Optionally, the execution logs can be visualized as a Gantt chart by
    `plotly.figure_factory.create_gantt`
    (https://plotly.github.io/plotly.py-docs/generated/plotly.figure_factory.create_gantt.html)
    if `plotly` is installed.
    """

    def __init__(
        self,
        gantt_filepath: str = None,
        gantt_params: Dict[str, Any] = {},
        metric_name_prefix: str = "_time_to_run ",
        task_name_func: Callable[[Node], str] = _get_task_name,
        time_log_filepath: str = None,
        enable_plotly: bool = True,
        enable_mlflow: bool = True,
    ):
        """
        Args:
            gantt_filepath: File path to save the generated gantt chart.
            gantt_params: Args fed to:
                https://plotly.github.io/plotly.py-docs/generated/plotly.figure_factory.create_gantt.html
            metric_name_prefix: Prefix for the metric names. The metric names are
                `metric_name_prefix` concatenated with the string returned by `task_name_func`.
            task_name_func: Callable to return the task name using ``kedro.pipeline.node.Node``
                object.
            time_log_filepath: File path to save the time log in JSON format.
            enable_plotly: Enable visualization of logged time as a gantt chart.
            enable_mlflow: Enable logging to MLflow.
        """
        self.enable_mlflow = find_spec("mlflow") and enable_mlflow
        self.enable_plotly = find_spec("plotly") and enable_plotly
        self.gantt_filepath = gantt_filepath
        self.gantt_params = gantt_params
        self.metric_name_prefix = metric_name_prefix
        self.task_name_func = task_name_func
        self.time_log_filepath = time_log_filepath or (
            tempfile.gettempdir() + "/_time_log.json"
        )
        Path(self.time_log_filepath).parent.mkdir(parents=True, exist_ok=True)
        dump_dict(
            self.time_log_filepath, {"time_begin": {}, "time_end": {}, "time": {}}
        )

        self._time_begin_dict = {}
        self._time_end_dict = {}
        self._time_dict = {}

    def update_time_dict(self, key: str, d: dict):
        dumping_dict = load_dict(self.time_log_filepath)
        dumping_dict[key].update(d)
        dump_dict(self.time_log_filepath, dumping_dict)

    def load_time_dict(self, key: str):
        return load_dict(self.time_log_filepath).get(key)

    @hook_impl
    def before_node_run(self, node, catalog, inputs):
        task_name = self.task_name_func(node)
        time_begin_dict = {task_name: time.time()}
        self._time_begin_dict.update(time_begin_dict)
        self.update_time_dict("time_begin", time_begin_dict)

    @hook_impl
    def after_node_run(self, node, catalog, inputs, outputs):
        task_name = self.task_name_func(node)

        time_end_dict = {task_name: time.time()}
        self._time_end_dict.update(time_end_dict)
        self.update_time_dict("time_end", time_end_dict)

        time_dict = {
            task_name: (
                self._time_end_dict.get(task_name)
                - self._time_begin_dict.get(task_name)
            )
        }

        log.info("Time duration: {}".format(time_dict))

        self._time_dict.update(time_dict)

        self.update_time_dict("time", time_dict)

        metric_time_dict = {
            (self.metric_name_prefix + k): v for (k, v) in time_dict.items()
        }
        mlflow_log_metrics(metric_time_dict, enable_mlflow=self.enable_mlflow)

    @hook_impl
    def after_pipeline_run(self, run_params, pipeline, catalog):

        self._time_begin_dict = self._time_begin_dict or self.load_time_dict(
            "time_begin"
        )
        self._time_end_dict = self._time_end_dict or self.load_time_dict("time_end")
        self._time_dict = self._time_dict or self.load_time_dict("time")

        log.info("Time duration: \n{}".format(pformat(self._time_dict)))

        if self.enable_plotly:
            if not (self._time_begin_dict and self._time_end_dict):
                log.warning(
                    "Time log dicts are not found. Skipping generating the Gantt Chart."
                )
                return

            tasks_reversed = list(self._time_begin_dict.keys())[::-1]

            from plotly.figure_factory import create_gantt

            df = [
                dict(
                    Task=t,
                    Start=self._time_begin_dict.get(t) * 1000,
                    Finish=self._time_end_dict.get(t) * 1000,
                )
                for t in tasks_reversed
            ]

            fig = create_gantt(df, **self.gantt_params)

            fp = self.gantt_filepath or (tempfile.gettempdir() + "/_gantt.html")
            Path(fp).parent.mkdir(parents=True, exist_ok=True)
            fig.write_html(fp)

            mlflow_log_artifacts(fp, enable_mlflow=self.enable_mlflow)



import copy
import numpy as np
from pathlib import Path

import torch
from torch.utils.data import DataLoader
import ignite
from ignite.engine import Events, create_supervised_trainer, create_supervised_evaluator
from ignite.metrics import RunningAverage
from ignite.handlers import EarlyStopping, ModelCheckpoint
from ignite.contrib.handlers.tqdm_logger import ProgressBar
from ignite.contrib.handlers.mlflow_logger import (
    MLflowLogger,
    OutputHandler,
    global_step_from_engine,
)
from ..handlers.time_limit import TimeLimit

import logging

log = logging.getLogger(__name__)

__all__ = ["NetworkTrain"]


class NetworkTrain:
    """Create a trainer for a supervised PyTorch model.

    Args:
        loss_fn (callable): Loss function used to train.
            Accepts an instance of loss functions at https://pytorch.org/docs/stable/nn.html#loss-functions
        epochs (int, optional): Max epochs to train
        seed (int, optional): Random seed for training.
        optimizer (torch.optim, optional): Optimizer used to train.
            Accepts optimizers at https://pytorch.org/docs/stable/optim.html
        optimizer_params (dict, optional): Parameters for optimizer.
        train_data_loader_params (dict, optional): Parameters for data loader for training.
            Accepts args at https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader
        val_data_loader_params (dict, optional): Parameters for data loader for validation.
            Accepts args at https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader
        evaluation_metrics (dict, optional): Metrics to compute for evaluation.
            Accepts dict of metrics at https://pytorch.org/ignite/metrics.html
        evaluate_train_data (str, optional): When to compute evaluation_metrics using training dataset.
            Accepts events at https://pytorch.org/ignite/engine.html#ignite.engine.Events
        evaluate_val_data (str, optional): When to compute evaluation_metrics using validation dataset.
            Accepts events at https://pytorch.org/ignite/engine.html#ignite.engine.Events
        progress_update (bool, optional): Whether to show progress bar using tqdm package
        scheduler (ignite.contrib.handle.param_scheduler.ParamScheduler, optional): Param scheduler.
            Accepts a ParamScheduler at
            https://pytorch.org/ignite/contrib/handlers.html#module-ignite.contrib.handlers.param_scheduler
        scheduler_params (dict, optional): Parameters for scheduler
        model_checkpoint (ignite.handlers.ModelCheckpoint, optional): Model Checkpoint.
            Accepts a ModelCheckpoint at https://pytorch.org/ignite/handlers.html#ignite.handlers.ModelCheckpoint
        model_checkpoint_params (dict, optional): Parameters for ModelCheckpoint at
            https://pytorch.org/ignite/handlers.html#ignite.handlers.ModelCheckpoint
        early_stopping_params (dict, optional): Parameters for EarlyStopping at
            https://pytorch.org/ignite/handlers.html#ignite.handlers.EarlyStopping
        time_limit (int, optioinal): Time limit for training in seconds.
        train_dataset_size_limit (int, optional): If specified, only the subset of training dataset is used.
            Useful for quick preliminary check before using the whole dataset.
        val_dataset_size_limit (int, optional): If specified, only the subset of validation dataset is used.
            useful for qucik preliminary check before using the whole dataset.
        cudnn_deterministic (bool, optional): Value for torch.backends.cudnn.deterministic.
            See https://pytorch.org/docs/stable/notes/randomness.html for details.
        cudnn_benchmark (bool, optional): Value for torch.backends.cudnn.benchmark.
            See https://pytorch.org/docs/stable/notes/randomness.html for details.
        mlflow_logging (bool, optional): If True and MLflow is installed, MLflow logging is enabled.

    Returns:
        trainer (callable): a callable to train a PyTorch model.

    """

    def __init__(
        self,
        loss_fn=None,
        epochs=None,
        seed=None,
        optimizer=None,
        optimizer_params=dict(),
        train_data_loader_params=dict(),
        val_data_loader_params=dict(),
        evaluation_metrics=None,
        evaluate_train_data=None,
        evaluate_val_data=None,
        progress_update=None,
        scheduler=None,
        scheduler_params=dict(),
        model_checkpoint=None,
        model_checkpoint_params=dict(),
        early_stopping_params=dict(),
        time_limit=None,
        train_dataset_size_limit=None,
        val_dataset_size_limit=None,
        cudnn_deterministic=None,
        cudnn_benchmark=None,
        mlflow_logging=True,
        train_params=dict(),
    ):
        self.train_params = dict(
            loss_fn=loss_fn,
            epochs=epochs,
            seed=seed,
            optimizer=optimizer,
            optimizer_params=optimizer_params,
            train_data_loader_params=train_data_loader_params,
            val_data_loader_params=val_data_loader_params,
            evaluation_metrics=evaluation_metrics,
            evaluate_train_data=evaluate_train_data,
            evaluate_val_data=evaluate_val_data,
            progress_update=progress_update,
            scheduler=scheduler,
            scheduler_params=scheduler_params,
            model_checkpoint=model_checkpoint,
            model_checkpoint_params=model_checkpoint_params,
            early_stopping_params=early_stopping_params,
            time_limit=time_limit,
            train_dataset_size_limit=train_dataset_size_limit,
            val_dataset_size_limit=val_dataset_size_limit,
            cudnn_deterministic=cudnn_deterministic,
            cudnn_benchmark=cudnn_benchmark,
        )
        self.train_params.update(train_params)
        self.mlflow_logging = mlflow_logging

    def __call__(self, model, train_dataset, val_dataset=None, **_):
        """Train a PyTorch model.

        Args:
            model (torch.nn.Module): PyTorch model to train.
            train_dataset (torch.utils.data.Dataset): Dataset used to train.
            val_dataset (torch.utils.data.Dataset, optional): Dataset used to validate.

        Returns:
            trained_model (torch.nn.Module): Trained PyTorch model.
        """
        assert train_dataset is not None
        train_params = self.train_params
        mlflow_logging = self.mlflow_logging

        if mlflow_logging:
            try:
                import mlflow  # NOQA
            except ImportError:
                log.warning("Failed to import mlflow. MLflow logging is disabled.")
                mlflow_logging = False

        loss_fn = train_params.get("loss_fn")
        assert loss_fn
        epochs = train_params.get("epochs")
        seed = train_params.get("seed")
        optimizer = train_params.get("optimizer")
        assert optimizer
        optimizer_params = train_params.get("optimizer_params", dict())
        train_dataset_size_limit = train_params.get("train_dataset_size_limit")
        if train_dataset_size_limit:
            train_dataset = PartialDataset(train_dataset, train_dataset_size_limit)
            log.info("train dataset size is set to {}".format(len(train_dataset)))

        val_dataset_size_limit = train_params.get("val_dataset_size_limit")
        if val_dataset_size_limit and (val_dataset is not None):
            val_dataset = PartialDataset(val_dataset, val_dataset_size_limit)
            log.info("val dataset size is set to {}".format(len(val_dataset)))

        train_data_loader_params = train_params.get("train_data_loader_params", dict())
        val_data_loader_params = train_params.get("val_data_loader_params", dict())
        evaluation_metrics = train_params.get("evaluation_metrics")
        evaluate_train_data = train_params.get("evaluate_train_data")
        evaluate_val_data = train_params.get("evaluate_val_data")
        progress_update = train_params.get("progress_update")

        scheduler = train_params.get("scheduler")
        scheduler_params = train_params.get("scheduler_params", dict())

        model_checkpoint = train_params.get("model_checkpoint")
        model_checkpoint_params = train_params.get("model_checkpoint_params")
        early_stopping_params = train_params.get("early_stopping_params")
        time_limit = train_params.get("time_limit")

        cudnn_deterministic = train_params.get("cudnn_deterministic")
        cudnn_benchmark = train_params.get("cudnn_benchmark")

        if seed:
            torch.manual_seed(seed)
            np.random.seed(seed)
        if cudnn_deterministic:
            torch.backends.cudnn.deterministic = cudnn_deterministic
        if cudnn_benchmark:
            torch.backends.cudnn.benchmark = cudnn_benchmark

        device = "cuda" if torch.cuda.is_available() else "cpu"
        model.to(device)
        optimizer_ = optimizer(model.parameters(), **optimizer_params)
        trainer = create_supervised_trainer(
            model, optimizer_, loss_fn=loss_fn, device=device
        )

        train_data_loader_params.setdefault("shuffle", True)
        train_data_loader_params.setdefault("drop_last", True)
        train_data_loader_params["batch_size"] = _clip_batch_size(
            train_data_loader_params.get("batch_size", 1), train_dataset, "train"
        )
        train_loader = DataLoader(train_dataset, **train_data_loader_params)

        RunningAverage(output_transform=lambda x: x, alpha=0.98).attach(
            trainer, "ema_loss"
        )

        RunningAverage(output_transform=lambda x: x, alpha=2 ** (-1022)).attach(
            trainer, "batch_loss"
        )

        if scheduler:

            class ParamSchedulerSavingAsMetric(
                ParamSchedulerSavingAsMetricMixIn, scheduler
            ):
                pass

            cycle_epochs = scheduler_params.pop("cycle_epochs", 1)
            scheduler_params.setdefault(
                "cycle_size", int(cycle_epochs * len(train_loader))
            )
            scheduler_params.setdefault("param_name", "lr")
            scheduler_ = ParamSchedulerSavingAsMetric(optimizer_, **scheduler_params)
            trainer.add_event_handler(Events.ITERATION_STARTED, scheduler_)

        if evaluate_train_data:
            evaluator_train = create_supervised_evaluator(
                model, metrics=evaluation_metrics, device=device
            )

        if evaluate_val_data:
            val_data_loader_params["batch_size"] = _clip_batch_size(
                val_data_loader_params.get("batch_size", 1), val_dataset, "val"
            )
            val_loader = DataLoader(val_dataset, **val_data_loader_params)
            evaluator_val = create_supervised_evaluator(
                model, metrics=evaluation_metrics, device=device
            )

        if model_checkpoint_params:
            assert isinstance(model_checkpoint_params, dict)
            minimize = model_checkpoint_params.pop("minimize", True)
            save_interval = model_checkpoint_params.get("save_interval", None)
            if not save_interval:
                model_checkpoint_params.setdefault(
                    "score_function", get_score_function("ema_loss", minimize=minimize)
                )
            model_checkpoint_params.setdefault("score_name", "ema_loss")
            mc = model_checkpoint(**model_checkpoint_params)
            trainer.add_event_handler(Events.EPOCH_COMPLETED, mc, {"model": model})

        if early_stopping_params:
            assert isinstance(early_stopping_params, dict)
            metric = early_stopping_params.pop("metric", None)
            assert (metric is None) or (metric in evaluation_metrics)
            minimize = early_stopping_params.pop("minimize", False)
            if metric:
                assert (
                    "score_function" not in early_stopping_params
                ), "Remove either 'metric' or 'score_function' from early_stopping_params: {}".format(
                    early_stopping_params
                )
                early_stopping_params["score_function"] = get_score_function(
                    metric, minimize=minimize
                )

            es = EarlyStopping(trainer=trainer, **early_stopping_params)
            if evaluate_val_data:
                evaluator_val.add_event_handler(Events.COMPLETED, es)
            elif evaluate_train_data:
                evaluator_train.add_event_handler(Events.COMPLETED, es)
            elif early_stopping_params:
                log.warning(
                    "Early Stopping is disabled because neither "
                    "evaluate_val_data nor evaluate_train_data is set True."
                )

        if time_limit:
            assert isinstance(time_limit, (int, float))
            tl = TimeLimit(limit_sec=time_limit)
            trainer.add_event_handler(Events.ITERATION_COMPLETED, tl)

        pbar = None
        if progress_update:
            if not isinstance(progress_update, dict):
                progress_update = dict()
            progress_update.setdefault("persist", True)
            progress_update.setdefault("desc", "")
            pbar = ProgressBar(**progress_update)
            pbar.attach(trainer, ["ema_loss"])

        else:

            def log_train_metrics(engine):
                log.info(
                    "[Epoch: {} | {}]".format(engine.state.epoch, engine.state.metrics)
                )

            trainer.add_event_handler(Events.EPOCH_COMPLETED, log_train_metrics)

        if evaluate_train_data:

            def log_evaluation_train_data(engine):
                evaluator_train.run(train_loader)
                train_report = _get_report_str(engine, evaluator_train, "Train Data")
                if pbar:
                    pbar.log_message(train_report)
                else:
                    log.info(train_report)

            eval_train_event = (
                Events[evaluate_train_data]
                if isinstance(evaluate_train_data, str)
                else Events.EPOCH_COMPLETED
            )
            trainer.add_event_handler(eval_train_event, log_evaluation_train_data)

        if evaluate_val_data:

            def log_evaluation_val_data(engine):
                evaluator_val.run(val_loader)
                val_report = _get_report_str(engine, evaluator_val, "Val Data")
                if pbar:
                    pbar.log_message(val_report)
                else:
                    log.info(val_report)

            eval_val_event = (
                Events[evaluate_val_data]
                if isinstance(evaluate_val_data, str)
                else Events.EPOCH_COMPLETED
            )
            trainer.add_event_handler(eval_val_event, log_evaluation_val_data)

        if mlflow_logging:
            mlflow_logger = MLflowLogger()

            logging_params = {
                "train_n_samples": len(train_dataset),
                "train_n_batches": len(train_loader),
                "optimizer": _name(optimizer),
                "loss_fn": _name(loss_fn),
                "pytorch_version": torch.__version__,
                "ignite_version": ignite.__version__,
            }
            logging_params.update(_loggable_dict(optimizer_params, "optimizer"))
            logging_params.update(_loggable_dict(train_data_loader_params, "train"))
            if scheduler:
                logging_params.update({"scheduler": _name(scheduler)})
                logging_params.update(_loggable_dict(scheduler_params, "scheduler"))

            if evaluate_val_data:
                logging_params.update(
                    {
                        "val_n_samples": len(val_dataset),
                        "val_n_batches": len(val_loader),
                    }
                )
                logging_params.update(_loggable_dict(val_data_loader_params, "val"))

            mlflow_logger.log_params(logging_params)

            batch_metric_names = ["batch_loss", "ema_loss"]
            if scheduler:
                batch_metric_names.append(scheduler_params.get("param_name"))

            mlflow_logger.attach(
                trainer,
                log_handler=OutputHandler(
                    tag="step",
                    metric_names=batch_metric_names,
                    global_step_transform=global_step_from_engine(trainer),
                ),
                event_name=Events.ITERATION_COMPLETED,
            )

            if evaluate_train_data:
                mlflow_logger.attach(
                    evaluator_train,
                    log_handler=OutputHandler(
                        tag="train",
                        metric_names=list(evaluation_metrics.keys()),
                        global_step_transform=global_step_from_engine(trainer),
                    ),
                    event_name=Events.COMPLETED,
                )
            if evaluate_val_data:
                mlflow_logger.attach(
                    evaluator_val,
                    log_handler=OutputHandler(
                        tag="val",
                        metric_names=list(evaluation_metrics.keys()),
                        global_step_transform=global_step_from_engine(trainer),
                    ),
                    event_name=Events.COMPLETED,
                )

        trainer.run(train_loader, max_epochs=epochs)

        try:
            if pbar and pbar.pbar:
                pbar.pbar.close()
        except Exception as e:
            log.error(e, exc_info=True)

        model = load_latest_model(model_checkpoint_params)(model)

        return model


def get_score_function(metric, minimize=False):
    def _score_function(engine):
        m = engine.state.metrics.get(metric)
        return -m if minimize else m

    return _score_function


def load_latest_model(model_checkpoint_params=None):
    if model_checkpoint_params and "model_checkpoint_params" in model_checkpoint_params:
        model_checkpoint_params = model_checkpoint_params.get("model_checkpoint_params")

    def _load_latest_model(model=None):
        if model_checkpoint_params:
            try:
                dirname = model_checkpoint_params.get("dirname")
                assert dirname
                dir_glob = Path(dirname).glob("*.pth")
                files = [str(p) for p in dir_glob if p.is_file()]
                if len(files) >= 1:
                    model_path = sorted(files)[-1]
                    log.info("Model path: {}".format(model_path))
                    loaded = torch.load(model_path)
                    save_as_state_dict = model_checkpoint_params.get(
                        "save_as_state_dict", True
                    )
                    if save_as_state_dict:
                        assert model
                        model.load_state_dict(loaded)
                    else:
                        model = loaded
                else:
                    log.warning("Model not found at: {}".format(dirname))
            except Exception as e:
                log.error(e, exc_info=True)
        return model

    return _load_latest_model


def _name(obj):
    return getattr(obj, "__name__", None) or getattr(obj.__class__, "__name__", "_")


def _clip_batch_size(batch_size, dataset, tag=""):
    dataset_size = len(dataset)
    if batch_size > dataset_size:
        log.warning(
            "[{}] batch size ({}) is clipped to dataset size ({})".format(
                tag, batch_size, dataset_size
            )
        )
        return dataset_size
    else:
        return batch_size


def _get_report_str(engine, evaluator, tag=""):
    report_str = "[Epoch: {} | {} | Metrics: {}]".format(
        engine.state.epoch, tag, evaluator.state.metrics
    )
    return report_str


def _loggable_dict(d, prefix=None):
    return {
        ("{}_{}".format(prefix, k) if prefix else k): (
            "{}".format(v) if isinstance(v, (tuple, list, dict, set)) else v
        )
        for k, v in d.items()
    }


class ParamSchedulerSavingAsMetricMixIn:
    """Base code:
    https://github.com/pytorch/ignite/blob/v0.2.1/ignite/contrib/handlers/param_scheduler.py#L49
    https://github.com/pytorch/ignite/blob/v0.2.1/ignite/contrib/handlers/param_scheduler.py#L163
    """

    def __call__(self, engine, name=None):

        if self.event_index != 0 and self.event_index % self.cycle_size == 0:
            self.event_index = 0
            self.cycle_size *= self.cycle_mult
            self.cycle += 1
            self.start_value *= self.start_value_mult
            self.end_value *= self.end_value_mult

        value = self.get_param()

        for param_group in self.optimizer_param_groups:
            param_group[self.param_name] = value

        if name is None:
            name = self.param_name

        if self.save_history:
            if not hasattr(engine.state, "param_history"):
                setattr(engine.state, "param_history", {})
            engine.state.param_history.setdefault(name, [])
            values = [pg[self.param_name] for pg in self.optimizer_param_groups]
            engine.state.param_history[name].append(values)

        self.event_index += 1

        if not hasattr(engine.state, "metrics"):
            setattr(engine.state, "metrics", {})
        engine.state.metrics[self.param_name] = value  # Save as a metric


class PartialDataset:
    def __init__(self, dataset, size):
        size = int(size)
        assert hasattr(dataset, "__getitem__")
        assert hasattr(dataset, "__len__")
        assert dataset.__len__() >= size
        self.dataset = dataset
        self.size = size

    def __len__(self):
        return self.size

    def __getitem__(self, item):
        return self.dataset[item]


class CopiedPartialDataset:
    def __init__(self, dataset, size):
        size = int(size)
        assert hasattr(dataset, "__getitem__")
        assert hasattr(dataset, "__len__")
        assert dataset.__len__() >= size
        self.dataset = [copy.deepcopy(dataset[i]) for i in range(size)]
        self.size = size

    def __len__(self):
        return self.size

    def __getitem__(self, item):
        return self.dataset[item]


class GetPartialDataset:
    def __init__(self, size):
        self.size = size

    def __call__(self, dataset):
        return CopiedPartialDataset(dataset, self.size)



import copy
from pathlib import Path
from typing import Any, Dict, List
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from ..core import AbstractVersionedDataSet, DataSetError, Version
import logging

log = logging.getLogger(__name__)


class SeabornPairPlotDataSet(AbstractVersionedDataSet):

    DEFAULT_SAVE_ARGS = dict()  # type: Dict[str, Any]

    def __init__(
        self,
        filepath: str,
        save_args: Dict[str, Any] = None,
        sample_args: Dict[str, Any] = None,
        version: Version = None,
    ) -> None:

        super().__init__(
            filepath=Path(filepath), version=version, exists_function=self._exists
        )
        self._load_args = {}
        self._save_args = save_args
        self._sample_args = sample_args

    def _describe(self) -> Dict[str, Any]:
        return dict(
            filepath=self._filepath,
            save_args=self._save_args,
            sampling_args=self._sample_args,
            version=self._version,
        )

    def _load(self) -> Any:
        """ loading is not supported. """
        return None

    def _save(self, data: pd.DataFrame) -> None:
        save_path = Path(self._get_save_path())
        save_path.parent.mkdir(parents=True, exist_ok=True)

        if self._sample_args is not None:
            data = data.sample(**self._sample_args)

        save_args = copy.deepcopy(self._save_args)

        cols_all = data.columns.to_list()

        chunk_size = save_args.pop("chunk_size", None)
        if chunk_size:
            x_vars = save_args.pop("x_vars", None)
            y_vars = save_args.pop("y_vars", None)

            p = Path(save_path)
            p.mkdir(parents=True, exist_ok=True)
            i = 0
            for x in _reshape(x_vars or cols_all, chunk_size):
                for y in _reshape(y_vars or cols_all, chunk_size):
                    log.info(
                        "Generating pairplot: x_vars = {}, y_vars = {}".format(x, y)
                    )
                    try:
                        plt.figure()
                        g = sns.pairplot(data, x_vars=x, y_vars=y, **save_args)
                        plt.suptitle("{} vs {}".format(y, x), va="bottom")
                        s = p / "{}_{:d}{}".format(p.stem, i, p.suffix)
                        i += 1
                        plt.savefig(s)
                        plt.close("all")
                    except Exception:
                        log.error(
                            "Failed to generate pairplot: x_vars = {}, y_vars = {}".format(
                                x, y
                            ),
                            exc_info=True,
                        )
        else:
            plt.figure()
            save_args.setdefault("x_vars", cols_all)
            save_args.setdefault("y_vars", cols_all)
            sns.pairplot(data, **self._save_args)
            plt.savefig(save_path)
            plt.close("all")

    def _exists(self) -> bool:
        try:
            path = self._get_load_path()
        except DataSetError:
            return False
        return Path(path).is_file()


def _reshape(ls: List[Any], size: int) -> List[List[Any]]:
    return [ls[i : i + size] for i in range(0, len(ls), size)]


# def _col_names_from_dtypes(
#     df: pd.DataFrame,
#     include: List[str] =["object"],
# ) -> List[str]:
#     dtypes_dict = df.dtypes.to_dict()
#     return [col for col, dtype in dtypes_dict.items() if dtype.name in include]
